@@ -10,6 +10,7 @@ from typing import Optional, Dict, Iterator, Union, List, Any, Tuple
 
 import requests
 import tabulator
+from ratelimit import sleep_and_retry, RateLimitDecorator
 from requests import Request
 from six.moves.urllib.parse import urlsplit, urlunsplit, parse_qsl, urlencode
 from tabulator.exceptions import TabulatorException
@@ -27,8 +28,14 @@ class DownloadError(Exception):
 
 class Download(object):
     """Download class with various download operations. Currently only GET requests are used and supported.
+    Requires either global user agent to be set or appropriate user agent parameter(s) to be completed.
 
     Args:
+        user_agent (Optional[str]): User agent string. HDXPythonUtilities/X.X.X- is prefixed.
+        user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
+        user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.
+        use_env (bool): Whether to read environment variables. Defaults to True.
+        rate_limit (Optional[Dict]): Rate limiting to use as a dict with calls and period. Defaults to None.
         **kwargs: See below
         auth (Tuple[str, str]): Authorisation information in tuple form (user, pass) OR
         basic_auth (str): Authorisation information in basic auth string form (Basic xxxxxxxxxxxxxxxx) OR
@@ -40,10 +47,17 @@ class Download(object):
         status_forcelist (List[int]): HTTP statuses for which to force retry
         method_whitelist (iterable): HTTP methods for which to force retry. Defaults t0 frozenset(['GET']).
     """
-    def __init__(self, **kwargs):
-        # type: (Any) -> None
-        self.session = get_session(**kwargs)
+
+    def __init__(self, user_agent=None, user_agent_config_yaml=None, user_agent_lookup=None, use_env=True,
+                 rate_limit=None, **kwargs):
+        # type: (Optional[str], Optional[str], Optional[str], bool, Optional[Dict], Any) -> None
+        self.session = get_session(user_agent, user_agent_config_yaml, user_agent_lookup, use_env, **kwargs)
         self.response = None
+        if rate_limit is not None:
+            self.setup = sleep_and_retry(
+                RateLimitDecorator(calls=rate_limit['calls'], period=rate_limit['period']).__call__(self.normal_setup))
+        else:
+            self.setup = self.normal_setup
 
     def close_response(self):
         # type: () -> None
@@ -167,7 +181,26 @@ class Download(object):
         full_url = urlunsplit(spliturl)
         return full_url, getparams
 
-    def setup(self, url, stream=True, post=False, parameters=None, timeout=None):
+    @staticmethod
+    def hxl_row(headers, hxltags, as_dict=False):
+        # type: (List[str], Dict[str,str], bool) -> Union[List[str], Dict[str,str]]
+        """Return HXL tag row for header row given list of headers and dictionary with header to HXL hashtag mappings.
+        Return list or dictionary depending upon teh as_dict argument.
+
+        Args:
+            headers (List[str]): Headers for which to get HXL hashtags
+            hxltags (Dict[str,str]): Header to HXL hashtag mapping
+
+        Returns:
+            Union[List[str],Dict[str,str]]: Return either a list or dictionary conating HXL hashtags
+
+        """
+        if as_dict:
+            return {header: hxltags.get(header, '') for header in headers}
+        else:
+            return [hxltags.get(header, '') for header in headers]
+
+    def normal_setup(self, url, stream=True, post=False, parameters=None, timeout=None):
         # type: (str, bool, bool, Optional[Dict], Optional[float]) -> requests.Response
         """Setup download from provided url returning the response
 
@@ -192,7 +225,7 @@ class Download(object):
                 self.response = self.session.get(self.get_url_for_get(url, parameters), stream=stream, timeout=timeout)
             self.response.raise_for_status()
         except Exception as e:
-            raisefrom(DownloadError, 'Setup of Streaming Download of %s failed!', e)
+            raisefrom(DownloadError, 'Setup of Streaming Download of %s failed!' % url, e)
         return self.response
 
     def hash_stream(self, url):
@@ -312,6 +345,8 @@ class Download(object):
         if file_type is not None:
             kwargs['format'] = file_type
             del kwargs['file_type']
+        if 'http_session' not in kwargs:
+            kwargs['http_session'] = self.session
         try:
             self.response = tabulator.Stream(url, **kwargs)
             self.response.open()
@@ -319,13 +354,12 @@ class Download(object):
         except TabulatorException as e:
             raisefrom(DownloadError, 'Getting tabular stream for %s failed!' % url, e)
 
-    def get_tabular_rows(self, url, dict_rows=False, **kwargs):
-        # type: (str, bool, Any) -> Iterator[Dict]
-        """Get iterator for reading rows from tabular data. Each row is returned as a dictionary.
+    def get_tabular_rows_as_list(self, url, **kwargs):
+        # type: (str, Any) -> Iterator[List]
+        """Get iterator for reading rows from tabular data. Each row is returned as a list.
 
         Args:
             url (str): URL to download
-            dict_rows (bool): Return dict (requires headers parameter) or list for each row. Defaults to False (list).
             **kwargs:
             headers (Union[int, List[int], List[str]]): Number of row(s) containing headers or list of headers
             file_type (Optional[str]): Type of file. Defaults to inferring.
@@ -335,7 +369,38 @@ class Download(object):
             Iterator[Union[List,Dict]]: Iterator where each row is returned as a list or dictionary.
 
         """
-        return self.get_tabular_stream(url, **kwargs).iter(keyed=dict_rows)
+        return self.get_tabular_stream(url, **kwargs).iter(keyed=False)
+
+    def get_tabular_rows(self, url, headers=1, dict_rows=False, insertions=None, **kwargs):
+        # type: (str, Union[int, List[int], List[str]], bool, Optional[Dict], Any) -> Tuple[List[str],Iterator[Union[List,Dict]]]
+        """Returns header of tabular file pointed to by url and an iterator where each row is returned as a list
+        or dictionary depending on the dict_rows argument. Optionally, headers and values can be inserted at specific
+        positions. This is achieved using the insertions argument. If supplied, it must be a dictionary containing
+        the keys "headers" and "functions". "headers" contains a list of tuples of the form (position, header) to be
+        inserted and "functions" is a list of functions each of which takes a parameter extended_rows which contains
+        the row's number, file headers list and the row values list.
+
+        Args:
+            url (str): URL to download
+            headers (Union[int, List[int], List[str]]): Number of row(s) containing headers or list of headers. Defaults to 1.
+            dict_rows (bool): Return dict (requires headers parameter) or list for each row. Defaults to False (list)
+            insertions (Optional[Dict]): Dictionary with additional headers and functions (see description).
+            **kwargs:
+            file_type (Optional[str]): Type of file. Defaults to inferring.
+            delimiter (Optional[str]): Delimiter used for values in each row. Defaults to inferring.
+
+        Returns:
+            Tuple[List[str],Iterator[Union[List,Dict]]]: Tuple (headers, iterator where each row is a list or dictionary)
+
+        """
+        if insertions is not None:
+            kwargs['post_parse'] = insertions['functions']
+        stream = self.get_tabular_stream(url, headers=headers, **kwargs)
+        headers = stream.headers
+        if insertions is not None:
+            for position, header in insertions['headers']:
+                headers.insert(position, header)
+        return headers, stream.iter(keyed=dict_rows)
 
     def download_tabular_key_value(self, url, **kwargs):
         # type: (str, Any) -> Dict
@@ -353,7 +418,7 @@ class Download(object):
 
         """
         output_dict = dict()
-        for row in self.get_tabular_rows(url, **kwargs):
+        for row in self.get_tabular_rows_as_list(url, **kwargs):
             if len(row) < 2:
                 continue
             output_dict[row[0]] = row[1]
@@ -425,3 +490,20 @@ class Download(object):
                     continue
                 output_dict[header][row[key_header]] = row[header]
         return output_dict
+
+    @staticmethod
+    def get_column_positions(headers):
+        # type: (List[str]) -> Dict[str,int]
+        """Get mapping of headers to column positions
+
+        Args:
+            headers (List[str]): List of headers
+
+        Returns:
+            Dict[str,int]: Dictionary where keys are header names and values are header positions
+
+        """
+        columnpositions = dict()
+        for i, header in enumerate(headers):
+            columnpositions[header] = i
+        return columnpositions
