@@ -155,7 +155,15 @@ class TestDownloader:
     def test_get_url_params_for_post(self):
         assert Download.get_url_params_for_post('http://www.lala.com/hdfa?a=3&b=4',
                                                 OrderedDict([('c', 'e'), ('d', 'f')])) == (
-               'http://www.lala.com/hdfa', OrderedDict([('a', '3'), ('b', '4'), ('c', 'e'), ('d', 'f')]))
+                   'http://www.lala.com/hdfa', OrderedDict([('a', '3'), ('b', '4'), ('c', 'e'), ('d', 'f')]))
+
+    def test_hxl_row(self):
+        headers = ['a', 'b', 'c']
+        hxltags = {'b': '#b', 'c': '#c'}
+        assert Download.hxl_row(headers, hxltags) == ['', '#b', '#c']
+        assert Download.hxl_row(headers, hxltags, as_dict=True) == {'a': '', 'b': '#b', 'c': '#c'}
+        assert Download.hxl_row(headers, dict()) == ['', '', '']
+        assert Download.hxl_row([], hxltags) == list()
 
     def test_setup_stream(self, fixtureurl, fixturenotexistsurl, getfixtureurl, postfixtureurl):
         with pytest.raises(DownloadError), Download() as downloader:
@@ -258,6 +266,61 @@ class TestDownloader:
             for y in result[x]:
                 result[x][y] = result[x][y].replace("'", '')
 
+    def test_get_tabular_rows_as_list(self, fixtureprocessurl):
+        with Download() as downloader:
+            rows = list(downloader.get_tabular_rows_as_list(fixtureprocessurl))
+            assert rows == [['la1', 'ha1', 'ba1', 'ma1'], ['header1', 'header2', 'header3', 'header4'],
+                            ['coal', '3', '7.4', 'needed'], ['gas', '2', '6.5', 'n/a']]
+
+    def test_get_tabular_rows(self, fixtureprocessurl):
+        with Download() as downloader:
+            expected = [['la1', 'ha1', 'ba1', 'ma1'], ['header1', 'header2', 'header3', 'header4'],
+                        ['coal', '3', '7.4', 'needed'], ['gas', '2', '6.5', 'n/a']]
+            expected_headers = expected[0]
+            headers, generator = downloader.get_tabular_rows(fixtureprocessurl)
+            assert headers == expected_headers
+            assert list(generator) == expected[1:]
+            headers, generator = downloader.get_tabular_rows(fixtureprocessurl, headers=1)
+            assert headers == expected_headers
+            assert list(generator) == expected[1:]
+            headers, generator = downloader.get_tabular_rows(fixtureprocessurl, headers=2)
+            assert headers == expected[1]
+            assert list(generator) == expected[2:]
+            headers, generator = downloader.get_tabular_rows(fixtureprocessurl, headers=[1, 2])
+            assert headers == ['la1 header1', 'ha1 header2', 'ba1 header3', 'ma1 header4']
+            assert list(generator) == expected[2:]
+            myheaders = ['a', 'b', 'c', 'd']
+            headers, generator = downloader.get_tabular_rows(fixtureprocessurl, headers=myheaders)
+            assert headers == myheaders
+            assert list(generator) == expected
+            headers, generator = downloader.get_tabular_rows(fixtureprocessurl, dict_rows=True, headers=1)
+            assert headers == expected_headers
+            expected_dicts = [{'la1': 'header1', 'ha1': 'header2', 'ba1': 'header3', 'ma1': 'header4'},
+                              {'la1': 'coal', 'ha1': '3', 'ba1': '7.4', 'ma1': 'needed'},
+                              {'la1': 'gas', 'ha1': '2', 'ba1': '6.5', 'ma1': 'n/a'}]
+            assert list(generator) == expected_dicts
+            headers, generator = downloader.get_tabular_rows(fixtureprocessurl, dict_rows=True, headers=3)
+            assert headers == expected[2]
+            expected_dicts = [{'coal': 'gas', '3': '2', '7.4': '6.5', 'needed': 'n/a'}]
+            assert list(generator) == expected_dicts
+
+            def testfn(extended_rows):
+                for row_number, headers, row in extended_rows:
+                    row.insert(2, 'lala')
+                    yield row_number, headers, row
+
+            insertions = {'headers': [(2, 'la')], 'functions': [testfn]}
+            headers, generator = downloader.get_tabular_rows(fixtureprocessurl, headers=3, insertions=insertions)
+            expected_headers_la = ['coal', '3', 'la', '7.4', 'needed']
+            assert headers == expected_headers_la
+            assert list(generator) == [['gas', '2', 'lala', '6.5', 'n/a']]
+
+            headers, generator = downloader.get_tabular_rows(fixtureprocessurl, dict_rows=True, headers=3,
+                                                             insertions=insertions)
+            assert headers == expected_headers_la
+            expected_dicts[0]['la'] = 'lala'
+            assert list(generator) == expected_dicts
+
     def test_download_tabular_rows_as_dicts(self, fixtureprocessurl):
         with Download() as downloader:
             result = downloader.download_tabular_rows_as_dicts(fixtureprocessurl, headers=2)
